/* Copyright (c) 1997-1999 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.

iemlib written by Thomas Musil (c) IEM KUG Graz Austria 2000 */


/* ---------- para_bp2~ - parametric bandpass 2. order ----------- */

typedef struct sigpara_bp2
{
    t_object x_obj;
    float    wn1;
    float    wn2;
    float    a0;
    float    a1;
    float    a2;
    float    b1;
    float    b2;
    float    sr;
    float    cur_f;
    float    cur_l;
    float    cur_a;
    float    cur_g;
    float    delta_f;
    float    delta_a;
    float    delta_g;
    float    end_f;
    float    end_a;
    float    end_g;
    float    ticks_per_interpol_time;
    float    rcp_ticks;
    float    interpol_time;
    int      ticks;
    int      counter_f;
    int      counter_a;
    int      counter_g;
    int      event_mask;
    float    x_msi;
} t_sigpara_bp2;

t_class *sigpara_bp2_class;

static void sigpara_bp2_calc(t_sigpara_bp2 *x)
{
    float l, al, gal, l2, rcp;

    l = x->cur_l;
    l2 = l*l + 1.0;
    al = l*x->cur_a;
    gal = al*x->cur_g;
    rcp = 1.0f/(al + l2);
    x->a0 = rcp*(l2 + gal);
    x->a1 = rcp*2.0f*(2.0f - l2);
    x->a2 = rcp*(l2 - gal);
    x->b1 = x->a1;
    x->b2 = rcp*(l2 - al);
}

static void sigpara_bp2_dsp_tick(t_sigpara_bp2 *x)
{
    if(x->event_mask)
    {
	if(x->counter_f)
	{
	    float l, si, co;

	    if(x->counter_f <= 1)
	    {
		x->cur_f = x->end_f;
		x->counter_f = 0;
		x->event_mask &= 6;/*set event_mask_bit 0 = 0*/
	    }
	    else
	    {
		x->counter_f--;
		x->cur_f *= x->delta_f;
	    }
	    l = x->cur_f * x->sr;
	    if(l <= 0.0f)
		x->cur_l = 1000000000000.0f;
	    else
	    {
		si = sin(l);
		co = cos(l);
		x->cur_l = co/si;
	    }
	}
	if(x->counter_a)
	{
	    if(x->counter_a <= 1)
	    {
		x->cur_a = x->end_a;
		x->counter_a = 0;
		x->event_mask &= 5;/*set event_mask_bit 1 = 0*/
	    }
	    else
	    {
		x->counter_a--;
		x->cur_a *= x->delta_a;
	    }
	}
	if(x->counter_g)
	{
	    if(x->counter_g <= 1)
	    {
		x->cur_g = x->end_g;
		x->counter_g = 0;
		x->event_mask &= 3;/*set event_mask_bit 2 = 0*/
	    }
	    else
	    {
		x->counter_g--;
		x->cur_g *= x->delta_g;
	    }
	}
	sigpara_bp2_calc(x);
    }
}

static t_int *sigpara_bp2_perform(t_int *w)
{
    float *in = (float *)(w[1]);
    float *out = (float *)(w[2]);
    t_sigpara_bp2 *x = (t_sigpara_bp2 *)(w[3]);
    int i, n = (t_int)(w[4]);
    float wn0, wn1=x->wn1, wn2=x->wn2;
    float a0=x->a0, a1=x->a1, a2=x->a2;
    float b1=x->b1, b2=x->b2;

    sigpara_bp2_dsp_tick(x);
    for(i=0; i<n; i++)
    {
	wn0 = *in++ - b1*wn1 - b2*wn2;
        *out++ = a0*wn0 + a1*wn1 + a2*wn2;
	wn2 = wn1;
	wn1 = wn0;
    }
    x->wn1 = wn1;
    x->wn2 = wn2;
    return(w+5);
}
/*   yn0 = *out;
     xn0 = *in;
     *************
     yn0 = a0*xn0 + a1*xn1 + a2*xn2 - b1*yn1 - b2*yn2;
     yn2 = yn1;
     yn1 = yn0;
     xn2 = xn1;
     xn1 = xn0;
     *************************
     y/x = (a0 + a1*z-1 + a2*z-2)/(1 + b1*z-1 + b2*z-2);*/

static void sigpara_bp2_ft4(t_sigpara_bp2 *x, t_floatarg t)
{
    int i = (int)((x->ticks_per_interpol_time)*t);

    x->interpol_time = t;
    if(i <= 0)
	i = 250;
    x->ticks = i;
    x->rcp_ticks = 1.0f / (float)i;
}

static void sigpara_bp2_ft3(t_sigpara_bp2 *x, t_floatarg l)
{
    float g = exp(0.11512925465 * l);

    if(g != x->cur_g)
    {
	x->end_g = g;
	x->counter_g = x->ticks;
	x->delta_g = exp(log(g/x->cur_g)*x->rcp_ticks);
	x->event_mask |= 4;/*set event_mask_bit 2 = 1*/
    }
}

static void sigpara_bp2_ft2(t_sigpara_bp2 *x, t_floatarg q)
{
    float a;

    if(q <= 0.0f)
	q = 0.000001f;
    a = 1.0f/q;
    if(a != x->cur_a)
    {
	x->end_a = a;
	x->counter_a = x->ticks;
	x->delta_a = exp(log(a/x->cur_a)*x->rcp_ticks);
	x->event_mask |= 2;/*set event_mask_bit 1 = 1*/
    }
}

static void sigpara_bp2_ft1(t_sigpara_bp2 *x, t_floatarg f)
{
    if(f <= 0.0f)
	f = 0.000001f;
    if(f != x->cur_f)
    {
	x->end_f = f;
	x->counter_f = x->ticks;
	x->delta_f = exp(log(f/x->cur_f)*x->rcp_ticks);
	x->event_mask |= 1;/*set event_mask_bit 0 = 1*/
    }
}

static void sigpara_bp2_dsp(t_sigpara_bp2 *x, t_signal **sp)
{
    float si, co, f;
    int i;

    x->sr = 3.14159265358979323846f / (float)(sp[0]->s_sr);
    x->ticks_per_interpol_time = 0.001f * (float)(sp[0]->s_sr) / (float)(sp[0]->s_n);
    i = (int)((x->ticks_per_interpol_time)*(x->interpol_time));
    if(i <= 0)
	i = 250;
    x->ticks = i;
    x->rcp_ticks = 1.0f / (float)i;
    f = (x->cur_f)*(x->sr);
    if(f <= 0.0f)
	f = 1000000000000.0f;
    else
    {
	si = sin(f);
	co = cos(f);
	f = co/si;
    }
    x->cur_l = f;
    dsp_add(sigpara_bp2_perform, 4, sp[0]->s_vec, sp[1]->s_vec, x, sp[0]->s_n);
}

static void *sigpara_bp2_new(t_symbol *s, int argc, t_atom *argv)
{
    t_sigpara_bp2 *x = (t_sigpara_bp2 *)pd_new(sigpara_bp2_class);
    int i;
    float si, co, f=0.0, q=1.0, l=0.0, interpol=0.0;

    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_float, gensym("ft1"));
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_float, gensym("ft2"));
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_float, gensym("ft3"));
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_float, gensym("ft4"));
    outlet_new(&x->x_obj, &s_signal);
    x->x_msi = 0;

    x->event_mask = 1;
    x->counter_f = 1;
    x->counter_a = 0;
    x->counter_g = 0;
    x->delta_f = 0.0f;
    x->delta_a = 0.0f;
    x->delta_g = 0.0f;
    x->interpol_time = 500.0f;
    x->wn1 = 0.0;
    x->wn2 = 0.0;
    x->a0 = 0.0;
    x->a1 = 0.0;
    x->a2 = 0.0;
    x->b1 = 0.0;
    x->b2 = 0.0;
    x->sr = 3.14159265358979323846f / 44100.0f;
    x->cur_a = 1.0;
    if((argc == 4)&&IS_A_FLOAT(argv,3)&&IS_A_FLOAT(argv,2)&&IS_A_FLOAT(argv,1)&&IS_A_FLOAT(argv,0))
    {
	f = (float)atom_getfloatarg(0, argc, argv);
        q = (float)atom_getfloatarg(1, argc, argv);
	l = (float)atom_getfloatarg(2, argc, argv);
	interpol = (float)atom_getfloatarg(3, argc, argv);
    }
    if(f <= 0.0f)
	f = 0.000001f;
    x->cur_f = f;
    f *= x->sr;
    if(f <= 0.0f)
	f = 1000000000000.0f;
    else
    {
	si = sin(f);
	co = cos(f);
	f = co/si;
    }
    x->cur_l = f;
    if(q <= 0.0f)
	q = 0.000001f;
    x->cur_a = 1.0f/q;
    x->cur_g = exp(0.11512925465 * l);
    if(interpol <= 0.0f)
	interpol = 500.0f;
    x->interpol_time = interpol;
    x->ticks_per_interpol_time = 0.5f;
    i = (int)((x->ticks_per_interpol_time)*(x->interpol_time));
    if(i <= 0)
	i = 250;
    x->ticks = i;
    x->rcp_ticks = 1.0f / (float)i;
    x->end_f = x->cur_f;
    x->end_a = x->cur_a;
    x->end_g = x->cur_g;
    return(x);
}

void sigpara_bp2_setup(void)
{
    sigpara_bp2_class = class_new(gensym("para_bp2~"), (t_newmethod)sigpara_bp2_new,
				0, sizeof(t_sigpara_bp2), 0, A_GIMME, 0);
    CLASS_MAINSIGNALIN(sigpara_bp2_class, t_sigpara_bp2, x_msi);
    class_addmethod(sigpara_bp2_class, (t_method)sigpara_bp2_dsp, gensym("dsp"), 0);
    class_addmethod(sigpara_bp2_class, (t_method)sigpara_bp2_ft1, gensym("ft1"), A_FLOAT, 0);
    class_addmethod(sigpara_bp2_class, (t_method)sigpara_bp2_ft2, gensym("ft2"), A_FLOAT, 0);
    class_addmethod(sigpara_bp2_class, (t_method)sigpara_bp2_ft3, gensym("ft3"), A_FLOAT, 0);
    class_addmethod(sigpara_bp2_class, (t_method)sigpara_bp2_ft4, gensym("ft4"), A_FLOAT, 0);
    class_sethelpsymbol(sigpara_bp2_class, gensym("iemhelp/help-para_bp2~"));
}
